//////////////////////////////////////////////////////////////////
// mysql JDBC Driver
//
// Version : 0.9.2
//
// JDBC : 1.22
//
// LastModified: 16.03.1997
//
// Copyright (c) 1997 GWE Technologies Ltd. All rights reserved
// See http://gwe.co.uk/mysql/jdbc/copyright.html for full details

package exgwe.sql;

import java.net.*;
import java.io.*;
import java.util.Hashtable;
import exgwe.sql.gweMysqlColumn;
import exgwe.sql.gweMysqlException;


// Class to read/write packets to the mysql server

class gweMysqlPacket {
	DataInputStream sin;
	DataOutputStream sout;
	byte packetSequence;
	int bpos, packetLength;
	public boolean TRACING=false;
	public final int HEADERLENGTH=4;
	public final int MAXBUFFER=65535;
	byte buffer[]=new byte[MAXBUFFER];

	// Set up given an input and output datastream
	public gweMysqlPacket(DataInputStream i, DataOutputStream o) {
		sin=i;
		sout=o;
	}

	// Set to -1; packetSequence gets ++'ed before sending
	void resetPacketSequence() {
		packetSequence=-1;
	}
	
	// Read in next packet; get upset if need be
	void next() throws gweMysqlException {		
		try {
			int i;
			byte[] b3 = new byte[3];
			
			// [3: packetLength] [packetSequence]
			for (i=0; i<3; i++) b3[i]=sin.readByte();
			packetLength=(int) (ub(b3[0])+(256*ub(b3[1]))+(256*256*ub(b3[2])));
			packetSequence=sin.readByte();

			if (TRACING) {
				System.out.println("Packet length = "+packetLength);
				System.out.println("Packet sequence = "+packetSequence);
			}
			
			// Get the data (block if necessary - thread higher up the us)
			sin.readFully(buffer, 0, packetLength);
			buffer[packetLength]=0;
			bpos=0;
			
			if (TRACING) {
				for (int j=0; j<packetLength; j++) {
					System.out.print(j+" : "+ub(buffer[j])+"  : ");
					if (buffer[j]>30) System.out.println((char)buffer[j]);
					else System.out.println("");
				}
			
			}
		} catch (Exception e) {throw new gweMysqlException("gweMysql.Next: "+e);}
	}

	
	// Return unsigned byte as int
	int ub(byte b) {
		return b<0?(int)(256+b):b;
	}

	// Where are we?
	int currentPos() {
		return bpos;
	}

	// Set position withtin packet data
	void setPos(int p) {
		bpos=p;
	}

	byte readByte() {
		return buffer[bpos++];
	}

	int read2Bytes() {
		return (int)(ub(buffer[bpos++])+(256*ub(buffer[bpos++])));
	}

	int read3Bytes() {
		return (int)(ub(buffer[bpos++])+(256*ub(buffer[bpos++]))+(256*256*ub(buffer[bpos++])));
	}

	long read4Bytes() {
		return (long)(ub(buffer[bpos++])+(256*ub(buffer[bpos++]))+(256*256*ub(buffer[bpos++]))+(256*256*256*ub(buffer[bpos++])));
	}


	// Read n bytes depending
	int readnBytes() {
		switch(ub(buffer[bpos++])) {
			case 1 : return ub(buffer[bpos++]);
			case 2 : return this.read2Bytes();
			case 3 : return this.read3Bytes();
			case 4 : return (int)this.read4Bytes();
			default : return 255;
		}
	}

	// read [:length]
	long readLength() {
		switch(ub(buffer[bpos])) {
			case 251 : {bpos++; return (long) 0; }
			case 252 : {bpos++; return (long) read2Bytes();}
			case 253 : {bpos++; return (long) read3Bytes();}
			case 254 : {bpos++; return (long) read4Bytes();}
			default  : return (long) ub(buffer[bpos++]);
		}
	}

	
	// Read null-terminated string
	String readString() {
		StringBuffer b = new StringBuffer();
		int i=bpos;
		while(buffer[i]!=0) b.append((char)buffer[i++]);
		bpos=i+1;
		return b.toString();
	}

	// Read given-length string
	String readLenString() {
		long len = this.readLength();
		if (len==0) return new String("");
		StringBuffer b = new StringBuffer((int)len);
		for (int i=0; i<len; i++) {
			if (buffer[bpos]==0) break;
			else b.append((char)buffer[bpos++]);
		}
		return b.toString();
	}

	// Is this the last packet?
	boolean isLastDataPacket() {
		return ((packetLength==1)&&(ub(buffer[0])==254));
	}

	// Reset position to start of packet data
	void clear() {
		bpos=HEADERLENGTH;
	}

	// Send packet
	void send() {
		try {
			int l=bpos;
			bpos=0;
			packetSequence++;
			this.write3Bytes(l-HEADERLENGTH);
			this.writeByte(packetSequence);

			if (TRACING) {
				for (int j=0; j<l; j++) {
					System.out.print(j+" : "+ub(buffer[j])+"  : ");
					if (buffer[j]>30) System.out.println((char)buffer[j]);
					else System.out.println("");
				}
			}

			sout.write(buffer, 0, l);
			sout.flush();
		} catch (Exception e) {System.out.println("send() Error : "+e);}
	}

	void writeByte(byte b) {
		buffer[bpos++]=b;
	}

	void write2Bytes(int i) {
		int r;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		buffer[bpos++]=(byte)i;
	}

	void write3Bytes(int i) {
		int r;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		buffer[bpos++]=(byte)i;
	}

	void write4Bytes(long i) {
		long r;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		r=i%256;
		buffer[bpos++]=(byte)r;
		i=i/256;
		buffer[bpos++]=(byte)i;
	}

	// Write null-terminated string
	void writeString(String s) {
		int l = s.length();
		for(int i=0; i<l; i++) buffer[bpos++]=(byte)s.charAt(i);
		buffer[bpos++]=0;
	}

	// Write string, with no termination
	void writeStringNoNull(String s) {
		int l = s.length();
		for(int i=0; i<l; i++) buffer[bpos++]=(byte)s.charAt(i);
	}

}


public class gweMysql {
	
	// Constants defined from mysql
	public static final int SLEEP=0;
	public static final int QUIT=1;
	public static final int INIT_DB=2;
	public static final int QUERY=3;
	public static final int FIELD_LIST=4;
	public static final int CREATE_DB=5;
	public static final int DROP_DB=6;
	public static final int RELOAD=7;
	public static final int SHUTDOWN=8;
	public static final int STATISTICS=9;
	public static final int PROCESS_INFO=10;
	public static final int CONNECT=11;
	public static final int FIELD_TYPE_DECIMAL=0;
	public static final int FIELD_TYPE_CHAR=1;
    public static final int FIELD_TYPE_SHORT=2;
	public static final int FIELD_TYPE_LONG=3;
    public static final int FIELD_TYPE_FLOAT=4;
	public static final int FIELD_TYPE_DOUBLE=5;
    public static final int FIELD_TYPE_NULL=6;
	public static final int FIELD_TYPE_TIME=7;
    public static final int FIELD_TYPE_LONGLONG=8;
	public static final int FIELD_TYPE_INT24=9;
    public static final int FIELD_TYPE_TINY_BLOB=249;
    public static final int FIELD_TYPE_MEDIUM_BLOB=250;
    public static final int FIELD_TYPE_LONG_BLOB=251;
    public static final int FIELD_TYPE_BLOB=252;
    public static final int FIELD_TYPE_VAR_STRING=253;
    public static final int FIELD_TYPE_STRING=254;

	
	Socket sock;
	gweMysqlPacket packet;
	String host;
	int port;
	String database;
	String user;
	String password;
	byte protocolVersion;
	String serverVersion;
	long threadId;
	String cryptSeed;
	long columnCount;
	gweMysqlColumn[] columnResult;
	Hashtable columnName;
	Hashtable columnFullName;
	int[] dataStart;

	public gweMysql() {}

	// Connect to mysql host		
	public synchronized void Connect(String h, int p, String d, String u, String pw) throws gweMysqlException {
		host=h;
		port=p;
		database=d;
		user=u;
		password=pw;
		if (user.equals("")) {user="nobody";}

		try {
			// Open TCP/IP socket
			sock = new Socket(host, port);
			DataInputStream sin = new DataInputStream(sock.getInputStream());
			DataOutputStream sout = new DataOutputStream(sock.getOutputStream());

			// Create new packet handler
			packet = new gweMysqlPacket(sin,sout);

//			packet.TRACING=true;

			// Get the first packet from mysql
			packet.next();
			protocolVersion = packet.readByte();
			serverVersion = packet.readString();
			long threadId = packet.read4Bytes();
			cryptSeed = packet.readString();

			// Send who we are
			packet.clear();
			packet.write2Bytes(0);
			packet.write3Bytes(packet.MAXBUFFER);
			packet.writeString(user);
			packet.writeString(cryptPassword(password,cryptSeed));
			packet.send();

			// Check if all is ok
			packet.next();
			byte statusCode = packet.readByte();
			if (statusCode==0xff) {
				String errorString = packet.readString();
				throw new gweMysqlException(errorString);
			} else if (statusCode==0x00) {
				long affectedRows = packet.readLength();
				long uniqueId = packet.readLength();
			} else throw new gweMysqlException("Unknown status code '"+statusCode+"'");

			// Ok, we're authenticated, so select our database
			packet.clear();
			packet.resetPacketSequence();
			packet.writeByte((byte)gweMysql.INIT_DB);
			packet.writeStringNoNull(database);
			packet.send();

			// Check if all is well
			packet.next();
			statusCode = packet.readByte();
			if (statusCode==0xff) {
				String errorString = packet.readString();
				throw new gweMysqlException(errorString);
			} else if (statusCode==0x00) {
				long affectedRows = packet.readLength();
				long uniqueId = packet.readLength();
			} else throw new gweMysqlException("Unknown status code '"+statusCode+"'");


		} catch (Exception e) {throw new gweMysqlException("gweMysql.Connect: "+e);}
	}

	// Return hash for given string
	long hashPassword(String p) {
		long  nr=1345345333;
		long  nr2=7;
		long tmp;
  		for (int i=0; i<p.length(); i++) {
			if ((p.charAt(i)==' ') || (p.charAt(i)=='\t') ) continue;
		    tmp=(long)p.charAt(i);
		    nr^= (((nr & 63)+nr2)*tmp)+ (nr << 8);
		    nr2+=tmp;
		  }
		return nr & (((long) 1L << 31) -1L); 
	}

	// Return encrypted response, given clear password and seed message
	String cryptPassword(String p, String m) {
		long hp, hm, s1, s2;
		long max_value= 0x01FFFFFF;
		double d;
		byte b;

		hp=this.hashPassword(p);
		hm=this.hashPassword(m);
		long nr=hp^hm;
		nr%=max_value;
		s1=nr;
		s2=nr/2;
		
		char[] cr = new char[m.length()];
		for (int i=0; i<m.length(); i++) {
			s1=(s1*3+s2)%max_value;
			s2=(s1+s2+33)%max_value;
			d = (double)s1/max_value;
			b = (byte)java.lang.Math.floor((d*31)+64);
			cr[i]=(char)b;
		}
		return new String(cr);
	}

	public gweMysql sqlQuery(String q) throws gweMysqlException {

		// Send query command and sql query string
		packet.clear();
		packet.resetPacketSequence();
		packet.writeByte((byte)gweMysql.QUERY);
		packet.writeStringNoNull(q);
		packet.send();

		// Check if we're ok
		packet.next();
		columnCount = packet.readLength();
		if (columnCount==255) {
			String errorString = packet.readString();
			throw new gweMysqlException(errorString);
		} else if (columnCount==0) {
			throw new gweMysqlException("No columns returned.");
		} else {
			columnResult = new gweMysqlColumn[(int)columnCount];
			columnName=new Hashtable((int)columnCount);
			columnFullName=new Hashtable((int)columnCount);
			dataStart = new int[(int)columnCount];

			// Read in the column information
			for (int i=0;i<columnCount; i++) {
				packet.next();
				String tableName=packet.readLenString();
				String colName=packet.readLenString();
				int colLength = packet.readnBytes();
				int colType=packet.readnBytes();
				packet.readByte(); // We know it's currently 2
				short colFlag = (short) packet.ub(packet.readByte());
				int colDecimals = packet.ub(packet.readByte());

				columnResult[i] = new gweMysqlColumn(tableName, colName, colLength, colType, colFlag, colDecimals);
				columnName.put(colName, new Integer(i));
				String fullName = tableName.concat(".").concat(colName);
				columnFullName.put(fullName, new Integer(i));

			}
			packet.next(); //Redundant - we know how many, this is end of section marker packet
		}
		return this;
	}


	public int sqlUpdate(String q) throws gweMysqlException {

		// Send query command and sql update
		packet.clear();
		packet.resetPacketSequence();
		packet.writeByte((byte)gweMysql.QUERY);
		packet.writeStringNoNull(q);
		packet.send();

		// check if we're ok
		packet.next();
		int updateStatus = packet.ub(packet.readByte());
		if (updateStatus==255) {
			String errorString = packet.readString();
			throw new gweMysqlException(errorString);
		}
		long affectedRows = packet.readLength();
		long updateId = packet.readLength();

		//return update id
		return (int)updateId;
	}


	public void sqlQuit() throws gweMysqlException {

		// Send quit command to mysql
		packet.clear();
		packet.resetPacketSequence();
		packet.writeByte((byte)gweMysql.QUIT);
		packet.send();

	}



	public boolean nextResult() throws gweMysqlException {
		int i, p;

		// Get the next incoming packet
		packet.next();
		// It it's not the last one, note start positions of data
		if(!packet.isLastDataPacket()) {
			for (i=0; i<columnCount; i++) {
				p=packet.currentPos();
				dataStart[i]=p;
				packet.setPos(1+p+(int)packet.readLength());
			}
			return true;
		} 
		return false;
	}

	public int getColumnCount() {
		return (int) columnCount;
	}
	
	public gweMysqlColumn getColumn(int c) throws gweMysqlException {
		c--;	// Columnd go from 1..columnCount to 0..columnCount-1
		if ((c<columnCount)&&(c>=0)) {
			return columnResult[c];
		} else {
			throw new gweMysqlException("Invalid column number : "+c);
		}
	}

	public boolean isNull(int c) throws gweMysqlException {
		c--;
		if ((c<columnCount)&&(c>=0)) {
			packet.setPos(dataStart[c]);
			return (packet.readLength()==0);
		} else throw new gweMysqlException("Invalid column number : "+c);
	}

	public int getColumnNum(String cn) throws gweMysqlException {
		Integer nI = (Integer) columnName.get(cn);
		if (nI==null) nI = (Integer) columnFullName.get(cn);
		if (nI!=null) return nI.intValue()+1;
		throw new gweMysqlException("Invalid column name : "+cn);
	}


	public String getString(int c) throws gweMysqlException {
		c--;
		if ((c<columnCount)&&(c>=0)) {
			packet.setPos(dataStart[c]);
			return packet.readLenString();
		} else throw new gweMysqlException("Invalid column number : "+c);
	}

	public String getString(String cn) throws gweMysqlException {
		Integer nI = (Integer) columnName.get(cn);
		if (nI==null) nI = (Integer) columnFullName.get(cn);
		if (nI!=null) {
			int c = nI.intValue()+1;
			if (c<columnCount) {
				packet.setPos(dataStart[c]);
				return packet.readLenString();
			}
		}
		throw new gweMysqlException("Invalid column string : "+cn);
	}

	public byte[] getBytes(int c) throws gweMysqlException {
		c--;
		if ((c<columnCount)&&(c>=0)) {
			packet.setPos(dataStart[c]);
			int l = (int)packet.readLength();
		    byte[] r = new byte[l];
			for (int i=0; i<l; i++) {
				r[i]=packet.readByte();
			}
			return r;
		}
		throw new gweMysqlException("Invalid column number : "+c);
	}


	public void reConnect() throws gweMysqlException {
		try {
			sqlQuit();
			sock.close();
			this.Connect(host, port, database, user, password);
		} catch (Exception e) {throw new gweMysqlException("gweMysql.reConnect: "+e);}
	}
		

	public void Close() throws gweMysqlException {
		try {
			sock.close();
		} catch (Exception e) {throw new gweMysqlException("gweMysql.Close: "+e);}
	}
}

